#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

#include "testfile.h"

#define notrace __attribute ((no_instrument_function))

int nlocks = 0;
void test_maxlock(int lock)
{
  if(lock>=nlocks)
    {
      nlocks = lock+1;
    }
}

notrace char *skip_spaces(char *tmp)
{
  while(isspace(*tmp) || *tmp=='\r' || *tmp=='\n')
    tmp++;
  return tmp;
}


void notrace make_job(taskjob_t *job, char **tmp)
{
  *tmp = skip_spaces(*tmp);
  if(!strncmp(*tmp,"+",1))
    {
      job->id = RUNNING;
      (*tmp)++;
    }
  else if(!strncmp(*tmp,"-",1))
    {
      job->id = IDLE;
      (*tmp)++;
    }
  else if(!strncmp(*tmp,"lockintr",8))
    {
      job->id = TLOCK_INTR;
      job->data.lock = strtol(*tmp+8,tmp,10);      
    }
  else if(!strncmp(*tmp,"lock",4))
    {
      job->id = TLOCK;
      job->data.lock = strtol(*tmp+4,tmp,10);      
    }
  else if(!strncmp(*tmp,"unlock",6))
    {
      job->id = UNLOCK;
      job->data.lock = strtol(*tmp+6,tmp,10);
    }
  else if(!strncmp(*tmp,"spinlock",8))
    {
      job->id = SPIN_LOCK;
      job->data.lock = strtol(*tmp+8,tmp,10);      
    }
  else if(!strncmp(*tmp,"spinunlock",10))
    {
      job->id = SPIN_UNLOCK;
      job->data.lock = strtol(*tmp+10,tmp,10);      
    }
  else if(!strncmp(*tmp,"signal",6))
    {
      job->id = SIGNAL;
      job->data.thread = strtol(*tmp+6,tmp,10);
    }
  else if(!strncmp(*tmp,"setstate",8))
    {
      job->id = SETSTATE;
      job->data.state = strtol(*tmp+8,tmp,10);
    }
  else
    {
      printf("Unknown job: |%s|\n",*tmp);
      exit(1);
    }

}

int totaltime = 0;
int nthreads = 0;
int priorities[MAX_TASKS];

notrace timeentry_t *make_timeentry(int line, char *buf)
{
  int i;
  char *tmp = buf;
  timeentry_t *res = calloc(1,sizeof(*res));;
  res->line = line;

  for(i=0;i<nthreads;i++)
    {
      make_job(&res->jobs[i],&tmp);
    }
  tmp = skip_spaces(tmp);
  if(*tmp)
    {
      res->timeout = strtol(tmp,&tmp,10);
      tmp = skip_spaces(tmp);
      if(*tmp)
	{
	  printf("line '%s' has too many charactes (%s)\n", buf,tmp);
	  exit(1);
	}
    }

  return res;
}


notrace void make_test(test_t *t, char **tmp)
{
  *tmp = skip_spaces(*tmp);
  if(!strncmp(*tmp,"?",1))
    {
      t->id = NOTEST;
      (*tmp)++;
    }
  else if(!strncmp(*tmp,"+",1))
    {
      t->id = NONBLOCKED;
      (*tmp)++;
    }
  else if(!strncmp(*tmp,"-",1))
    {
      t->id = BLOCKED;
      (*tmp)++;
    }
  else if(!strncmp(*tmp,"prio",4))
    {
      t->id = PRIORITY;
      t->data.priority = strtol(*tmp+4,tmp,10);      
    }
  else if(!strncmp(*tmp,"state",5))
    {
      t->id = STATE;
      t->data.state = strtol(*tmp+5,tmp,10);      
    }
  else if(!strncmp(*tmp,"lockcount",9))
    {
      t->id = LOCKCOUNT;
      t->data.state = strtol(*tmp+9,tmp,10);      
    }
  else
    {
      printf("Unknown test: |%s|\n",*tmp);
      exit(1);
    }

}

notrace void make_testentry(int linenum, timeentry_t *te, char *line)
{
  int i;
  testentry_t **add;

  for(add = &te->tests; *add; add=&((*add)->next) )
    ;
  
  *add = calloc(1,sizeof(testentry_t));
  (*add)->line = linenum;

  for(i=0; i<nthreads;i++)
    {
      make_test(&(*add)->tests[i],&line);
    }
}

timeentry_t *first_time = NULL;

void notrace read_testfile(char *filename)
{
  int line = 0;
  char buf[1000];
  timeentry_t *prev;

  FILE *fp = fopen(filename,"r");
  if(!fp)
    {
      printf("Can't open %s\n", filename);
      exit(1);
    }
  
  while(fgets(buf,1000,fp))
  {
    char *tmp = skip_spaces(buf);
    line++;

    if(!*tmp || *tmp == '#')
      {
	continue;
      }
    if(!strncmp(tmp,"threads:",8))
      {
	nthreads = 0;
	tmp += 8;
	while(*tmp)
	  {
	    priorities[nthreads] = strtol(tmp,&tmp,10);
	    nthreads++;
	    tmp = skip_spaces(tmp);
	  }
	continue;
      }
    else
      {
	if(nthreads<1)
	  {
	    printf("You must start with threads: <priorities>\n");
	    exit(1);
	  }
      }


    if(!strncmp(tmp,"test:",5))
      {
	if(!first_time)
	  {
	    printf("You can't start with a test\n");
	  exit(1);
	  }
	make_testentry(line,prev,tmp+5);
	continue;
      }
    if(!first_time)
      {
	first_time = make_timeentry(line,buf);
	prev = first_time;
      }
    else
      {
	prev->next = make_timeentry(line,buf);
	prev = prev->next;
      }    
  }
  fclose(fp);
}
