#include "testtask.h"
#include "osfuncs.h"
#include "linux/init_task.h"

/* #define DEBUG printf */
#define DEBUG(...) 

int initialized = 0;
pthread_mutex_t global_lock;

testtask_t *current_task = NULL;


void notrace setup_current(testtask_t *task)
{
  memcpy(&___current,&task->task,sizeof(task->task));
  memcpy(&current_ti,&task->thread,sizeof(task->thread));
  current_task = task;
  irqflags = task->irqflags;
}


void notrace switch2testtask(testtask_t *task)
{
  testtask_t *old;
  old = current_task;
  if(task==old)
    {
      return;
    }
  current_task = task;
  if(pthread_cond_signal(&task->cond)!=0)
    {
      printf("pthread_cond_signal failed\n");
      exit(1);
    }
  DEBUG("%p signaled %p\n", old, task);
  if(old)
    {
      old->irqflags = irqflags;
      DEBUG("%p waiting\n",old);
      if(pthread_cond_wait(&old->cond, &global_lock)!=0)
	{
	  printf("pthread_cond_wait failed\n");
	  exit(1);
	}
      DEBUG("%p awoken\n",old);
      setup_current(old);
    }

}

notrace void *testtask_entry(void *arg)
{
  testtask_t *self = arg;
  DEBUG("%p awake\n",self);
  pthread_mutex_lock(&global_lock);
  DEBUG("%p got lock\n",self);
  while(self->isalive)
    {
      DEBUG("%p waiting\n", self);
      pthread_cond_wait(&self->cond,&global_lock);
      DEBUG("%p awoken\n",self);
      setup_current(self);
      testtask_job();
    }
  pthread_mutex_lock(&global_lock);
  return NULL;
}

notrace testtask_t *create_testtask(int pid,int prio)
{
  pthread_attr_t attr;
  testtask_t *task = calloc(1,sizeof(testtask_t));
  
  if(!initialized)
    {
      pthread_mutex_init(&global_lock,NULL);
      initialized = 1;
    }

  pthread_cond_init(&task->cond,NULL);
  task->task = calloc(1,sizeof(*task->task));
  spin_lock_init(&task->task->pi_lock);
  plist_head_init(&task->task->pi_waiters);
  
  task->thread = calloc(1,sizeof(struct thread_info));
  task->thread->task = task->task;
  task->task->thread_info = task->thread;

  task->task->pid = pid;
  task->task->normal_prio = prio;
  task->task->prio = prio;
  task->task->lock_depth = -1;		/* -1 = no lock */

  task->isalive = 1;

  if(pthread_attr_init(&attr))
    {
      printf("pthread_attr_init failed\n");
      exit(1);
    }
  if(pthread_create(&task->pthread, &attr, 
		    testtask_entry, task))
  {
    printf("pthread_create failed\n");
    exit(1);
  }
  else
    {
      DEBUG("pthread_create ok\n");
    }
  

  return task;
}

static struct fs_struct init_fs; /* = INIT_FS(init_fs); */
static struct files_struct init_files;/*  = INIT_FILES(init_files); */
static struct signal_struct init_signals;/*  = INIT_SIGNALS(init_signals); */
static struct sighand_struct init_sighand;/*  = INIT_SIGHAND(init_sighand); */
struct mm_struct init_mm;/*  = INIT_MM(init_mm); */
union thread_union init_thread_union;
struct group_info init_groups;
struct user_struct root_user;
struct task_struct init_task = INIT_TASK(init_task);
